/**
 * Copyright (c) 2010 Alois Glomann gnalff@lucubrators.net
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.lucubrators.honeycomb.xml;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.lucubrators.honeycomb.core.Comb;
import net.lucubrators.honeycomb.core.constraint.ConstraintHolder;
import net.lucubrators.honeycomb.xml.constraint.ConstraintProvider;
import net.lucubrators.honeycomb.xml.controller.ControllerProvider;
import net.lucubrators.honeycomb.xml.exceptions.XMLParsingException;
import net.lucubrators.honeycomb.xml.registry.CombRegistry;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.filter.ElementFilter;
import org.jdom.filter.Filter;
import org.jdom.input.SAXBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Parses an xml containing the combs written in the markup descriped by the
 * honeycomb.xsd (in the META-INF folder)
 * 
 * @author gnalFF
 * 
 */
@SuppressWarnings("serial")
public class HoneycombXMLParser {

	private static Logger		log					= LoggerFactory.getLogger(HoneycombXMLParser.class);

	public static final String	COMB_TAG			= "comb";
	public static final String	MARK_TAG			= "mark";
	public static final String	ATTR_TAG			= "attr";

	public static final String	ID_ATTR				= "id";
	public static final String	PARENT_ATTR			= "parent";
	public static final String	CONTROLLER_REF		= "controllerRef";
	public static final String	EVENT_NAME			= "eventName";
	public static final String	KEY_ATTR			= "key";
	public static final String	VALUE_ATTR			= "value";
	public static final String	NAME_ATTR			= "name";

	/**
	 * Returns only Elements with id attribute and no childs. Use only in
	 * conjunction with {@link ElementFilter}
	 */
	private static final Filter	ROOT_FILTER			= new Filter() {

														@Override
														public boolean matches(final Object obj) {
															Element comb = (Element) obj;
															String parent = comb.getAttributeValue(PARENT_ATTR);
															if (null != parent) {
																return false;
															}
															@SuppressWarnings("rawtypes")
															List children = comb.getChildren(MARK_TAG, comb.getNamespace());
															if (children.size() > 0) {
																return false;
															}
															return true;
														}
													};
	/**
	 * Returns only Elements which have a parent attribute. Use only in
	 * conjunction with {@link ElementFilter}
	 */
	private static final Filter	WITH_CHILD_FILTER	= new Filter() {

														@Override
														public boolean matches(final Object obj) {
															Element comb = (Element) obj;
															String parent = comb.getAttributeValue(PARENT_ATTR);
															return null != parent || comb.getChildren(MARK_TAG, comb.getNamespace()).size() > 0;
														}
													};

	private static final String	CONSTRAINT_TAG		= "constraint";

	private CombRegistry		registry;

	private Document			document;
	private Element				rootElement;

	private ControllerProvider	controllerProvider;
	private ConstraintProvider	constraintProvider;

	/**
	 * Constructor
	 * 
	 * @param toScan
	 *            the file which should be scanned
	 */
	public HoneycombXMLParser(final File toScan) {
		SAXBuilder saxBuilder = new SAXBuilder();
		try {
			document = saxBuilder.build(toScan);
		} catch (JDOMException e) {
			throw new XMLParsingException("Could not build document from " + toScan.getAbsolutePath(), e);
		} catch (IOException e) {
			throw new XMLParsingException("Could not build document from " + toScan.getAbsolutePath(), e);
		}
		rootElement = document.getRootElement();
	}

	/**
	 * Starts the scanning process.
	 */
	@SuppressWarnings("unchecked")
	public void scan() {
		log.trace("Scanning for controllers using controllerprovider: {}", controllerProvider.getClass().getName());

		log.trace("Scanning for combs");
		List<Element> rootCombs = rootElement.getContent(new ElementFilter(COMB_TAG).and(ROOT_FILTER));
		for (Element element : rootCombs) {
			Comb createComb = createComb(element);
			registry.register(createComb);
		}
		List<Element> combsWithChild = rootElement.getContent(new ElementFilter(COMB_TAG).and(WITH_CHILD_FILTER));
		for (Element element : combsWithChild) {
			Comb createComb = createComb(element);
			registry.register(createComb);
		}
	}

	private Comb createComb(final Element element) {
		String id = element.getAttributeValue(ID_ATTR);
		String parent = element.getAttributeValue(PARENT_ATTR);
		String controller = element.getAttributeValue(CONTROLLER_REF);
		String eventName = element.getAttributeValue(EVENT_NAME);
		Comb ret = null;
		if (null != id && parent == null) {
			log.trace("creating new comb with id {} and controllerRef {}.", id, controller);
			ret = new Comb(id, controllerProvider.getController(controller));
		} else if (id != null && parent != null) {
			log.trace("creating new alias comb with id {} from parent {}.", id, parent);
			Comb parentComb = registry.getById(parent);
			ret = parentComb.alias(id);
		} else if (null == id && null != parent) {
			log.trace("creating copy comb from parent {}.", id, parent);
			Comb parentComb = registry.getById(parent);
			ret = parentComb.copy();
		} else {
			throw new XMLParsingException("Not allowed combination of id " + id + " and parent " + parent);
		}
		Map<String, String> attributes = createAttributes(element);
		ret.setAttributes(attributes);
		if (eventName != null) {
			ret.setEventName(eventName);
		}
		scanMarksAndAdd(element, ret);
		ArrayList<ConstraintHolder> constraints = scanConstraints(element);
		if (!constraints.isEmpty()) {
			ret.setConstraintHolders(constraints);
		}
		return ret;
	}

	@SuppressWarnings("unchecked")
	private ArrayList<ConstraintHolder> scanConstraints(final Element element) {
		ArrayList<ConstraintHolder> constraintHolders = new ArrayList<ConstraintHolder>();
		List<Element> children = element.getChildren(CONSTRAINT_TAG, element.getNamespace());
		for (Element constraint : children) {
			String constraintParent = constraint.getAttributeValue(PARENT_ATTR);
			if (null == constraintParent || constraintParent.isEmpty()) {
				throw new IllegalArgumentException("No constraint parent specified");
			}
			Object object = constraintProvider.getConstraint(constraintParent);
			constraintHolders.add(new ConstraintHolder(object, createAttributes(constraint)));
		}
		return constraintHolders;
	}

	@SuppressWarnings("unchecked")
	private void scanMarksAndAdd(final Element element, final Comb toAdd) {
		List<Element> marks = element.getChildren(MARK_TAG, element.getNamespace());
		HashMap<String, List<Comb>> marksMap = new HashMap<String, List<Comb>>();
		for (Element mark : marks) {
			String markName = mark.getAttributeValue(NAME_ATTR);
			ArrayList<Comb> markChildren = new ArrayList<Comb>();
			List<Element> children = mark.getChildren(COMB_TAG, mark.getNamespace());
			for (Element child : children) {
				Comb childComb = createComb(child);
				markChildren.add(childComb);
			}
			marksMap.put(markName, markChildren);
		}
		toAdd.setMarks(marksMap);
	}

	@SuppressWarnings("unchecked")
	private Map<String, String> createAttributes(final Element element) {
		HashMap<String, String> ret = new HashMap<String, String>();
		List<Element> children = element.getChildren(ATTR_TAG, rootElement.getNamespace());
		for (Element attribute : children) {
			String key = attribute.getAttributeValue(KEY_ATTR);
			String value = attribute.getAttributeValue(VALUE_ATTR);
			ret.put(key, value);
		}
		return ret;
	}

	public void setRegistry(final CombRegistry registry) {
		this.registry = registry;
	}

	public void setControllerProvider(final ControllerProvider controllerProvider) {
		this.controllerProvider = controllerProvider;
	}

	public void setConstraintProvider(final ConstraintProvider constraintProvider) {
		this.constraintProvider = constraintProvider;
	}

}
